from __future__ import division
import pandas as pd
from pyteomics import pepxml, achrom, auxiliary as aux, mass, fasta, mzid, parser
import numpy as np
import random
from catboost import CatBoostClassifier
from sklearn.model_selection import train_test_split
import os
from collections import Counter, defaultdict
from .utils_figures import get_fdbinsize
from scipy.stats import scoreatpercentile
from sklearn.isotonic import IsotonicRegression
import logging
import warnings
import re
warnings.formatwarning = lambda msg, *args, **kw: str(msg) + '\n'
logger = logging.getLogger(__name__)
SEED = 42


class NoDecoyError(ValueError):
    pass


class WrongInputError(NotImplementedError):
    pass


class EmptyFileError(ValueError):
    pass


def filter_custom(df, fdr, key, is_decoy, reverse, remove_decoy, ratio, formula, correction=None, loglabel=None):
    kw = dict(key=key, is_decoy=is_decoy, reverse=reverse, full_output=True,
        remove_decoy=False, ratio=ratio, formula=formula)
    df = df.copy()
    q = aux.qvalues(df, correction=1, **kw)
    q_uncorr = aux.qvalues(df, correction=0, **kw)
    df['q'] = q['q']
    df['q_uncorrected'] = q_uncorr['q']

    if correction is not None:
        qlabel = 'q' if correction else 'q_uncorrected'
        logger.debug('Explicitly using %s for filtering.', qlabel)
    elif df['q'].min() < fdr:
        logger.debug('Successfully filtered with +1 correction (label = %s).', loglabel)
        qlabel = 'q'
    else:
        logger.info('No results for filtering with +1 correction (label = %s). Rerunning without correction...', loglabel)
        qlabel = 'q_uncorrected'
    if remove_decoy:
        df = df[~df[is_decoy]]
    return df[df[qlabel] < fdr].copy()


def convert_tandem_cleave_rule_to_regexp(cleavage_rule):

    def get_sense(c_term_rule, n_term_rule):
        if '{' in c_term_rule:
            return 'N'
        elif '{' in n_term_rule:
            return 'C'
        else:
            if len(c_term_rule) <= len(n_term_rule):
                return 'C'
            else:
                return 'N'

    def get_cut(cut, no_cut):
        aminoacids = set(parser.std_amino_acids)
        cut = ''.join(aminoacids & set(cut))
        if '{' in no_cut:
            no_cut = ''.join(aminoacids & set(no_cut))
            return cut, no_cut
        else:
            no_cut = ''.join(set(parser.std_amino_acids) - set(no_cut))
            return cut, no_cut

    out_rules = []
    for protease in cleavage_rule.split(','):
        protease = protease.replace('X', ''.join(parser.std_amino_acids))
        c_term_rule, n_term_rule = protease.split('|')
        sense = get_sense(c_term_rule, n_term_rule)
        if sense == 'C':
            cut, no_cut = get_cut(c_term_rule, n_term_rule)
        else:
            cut, no_cut = get_cut(n_term_rule, c_term_rule)

        if no_cut:
            if sense == 'C':
                out_rules.append('([%s](?=[^%s]))' % (cut, no_cut))
            else:
                out_rules.append('([^%s](?=[%s]))' % (no_cut, cut))
        else:
            if sense == 'C':
                out_rules.append('([%s])' % (cut, ))
            else:
                out_rules.append('(?=[%s])' % (cut, ))
    return '|'.join(out_rules)


def calc_TOP3(df):
    df['TOP3'] = df['TOP3'].apply(lambda z: sum(sorted(z, reverse=True)[:3]))


def calc_NSAF(df):
    df['NSAF'] = df['PSMs'] / df['length']
    NSAF_sum = np.sum(df['NSAF'])
    df['NSAF'] = df['NSAF'] / NSAF_sum
    if sum(pd.notna(df['NSAF'])):
        df['LOG10_NSAF'] = np.log10(df['NSAF'])
    df.loc[pd.isna(df['NSAF']), 'NSAF'] = 0.0


def keywithmaxval(d):
    # this method is much faster than using max(prots.iterkeys(), key=(lambda key: prots[key]))
    v = list(d.values())
    k = list(d.keys())
    return k[v.index(max(v))]


def add_protein_groups(df, df_ms1_path):
    pept_prots = defaultdict(set)
    prot_prots = defaultdict(set)
    prot_pepts = dict()
    for peptides, dbname in df.sample(frac=1).reset_index(drop=True)[['peptides set', 'dbname']].values:
        prot_pepts[dbname] = peptides
        for peptide in peptides:
            pept_prots[peptide].add(dbname)
    for prots in pept_prots.values():
        for dbname in prots:
            prot_prots[dbname].update(prots)
    prot_pepts_count = dict()
    prot_pepts_count2 = dict()
    
    if not df_ms1_path:
        for k, v in prot_pepts.items():
            prot_pepts_count[k] = len(v)
            prot_pepts_count2[k] = len(v)

    else:
        df_ms1 = pd.read_csv(df_ms1_path, sep='\t')
        ms1s = dict()
        for qval, prot in df_ms1[['score', 'dbname']].values:
            ms1s[prot] = float(qval)

        max_k = max(ms1s.values())
        for k, v in prot_pepts.items():
            prot_pepts_count[k] = len(v) + ms1s.get(k, 0) / max_k
            prot_pepts_count2[k] = len(v)
    tostay = set()
    while pept_prots:
        bestprot = keywithmaxval(prot_pepts_count)
        tostay.add(bestprot)
        for pep in prot_pepts[bestprot]:
            for k in pept_prots[pep]:
                prot_pepts_count[k] -= 1
                prot_pepts_count2[k] -= 1
            del pept_prots[pep]
            
        for k, v in list(prot_pepts_count2.items()):
            if v == 0:
                del prot_pepts_count[k]
                del prot_pepts_count2[k]
    df['groupleader'] = df['dbname'].apply(lambda x: x in tostay)
    df['all proteins'] = df['dbname'].apply(lambda x: ';'.join(prot_prots[x]))


def process_fasta(df, path_to_fasta, decoy_prefix, decoy_infix=False):
    protsS = dict()
    decoy_check_flag = False
    for x in fasta.read(path_to_fasta):
        dbname = x[0].split(' ')[0]
        if not decoy_check_flag:
            if (not decoy_infix and dbname.startswith(decoy_prefix)) or (decoy_infix and decoy_infix in dbname):
                decoy_check_flag = True
        protsS[dbname] = x[1]
    df['sequence'] = df['dbname'].apply(lambda x: protsS.get(x, protsS.get(x.split(' ')[0], '')))
    if not decoy_check_flag:
        if not decoy_infix:
            df['sequence'] = df.apply(
                lambda x: x['sequence'] if x['sequence'] else protsS.get(
                    x['dbname'].replace(decoy_prefix, ''), protsS.get(x['dbname'].split(' ')[0].replace(decoy_prefix, ''), '')),
                axis=1)
        else:
            df['sequence'] = df.apply(
                lambda x: x['sequence'] if x['sequence'] else protsS.get(
                    x['dbname'].replace(decoy_infix, ''), protsS.get(x['dbname'].split(' ')[0].replace(decoy_infix, ''), '')),
                axis=1)
    return df


def get_tag_names(columns):
    return [c for c in columns if c[:4] == 'tag_']


def get_proteins_dataframe(df1_f2, decoy_prefix, all_decoys_2, decoy_infix=False, path_to_fasta=False, pif_threshold=0):
    proteins_dict = dict()
    cols = ['protein', 'protein_descr', 'peptide', 'PEP', 'MS1Intensity', 'PIF']
    tagnames = get_tag_names(df1_f2.columns)
    tagsums = []
    for c in tagnames:
        cols.append(c)
        tagsums.append(0.)
    have_pif = 'PIF' in df1_f2.columns
    if not have_pif:
        logger.debug('PIF not found.')
        if pif_threshold > 0:
            logger.warning('PIF not found, threshold will not be applied.')
        df1_f2['PIF'] = 0.

    for tup in df1_f2[cols].values:
        # avoid star unpacking to support Python 2.7 for a little longer
        (proteins, protein_descriptions, peptide, pep, ms1_i, pif), tags = tup[:6], tup[6:]
        for prot, prot_descr in zip(proteins, protein_descriptions):
            if prot not in proteins_dict:
                proteins_dict[prot] = dict()
                proteins_dict[prot]['dbname'] = prot
                proteins_dict[prot]['description'] = prot_descr
                proteins_dict[prot]['PSMs'] = 0
                proteins_dict[prot]['peptides set'] = set()
                proteins_dict[prot]['sequence'] = ''
                proteins_dict[prot]['NSAF'] = 0
                proteins_dict[prot]['TOP3'] = []
                proteins_dict[prot]['sq'] = 0
                proteins_dict[prot]['score'] = dict()
                proteins_dict[prot]['q-value'] = 1.0
                for tag in tagnames:
                    proteins_dict[prot][tag] = 0.
                if not decoy_infix:
                    proteins_dict[prot]['decoy'] = prot.startswith(decoy_prefix)
                else:
                    proteins_dict[prot]['decoy'] = decoy_infix in prot
                proteins_dict[prot]['decoy2'] = prot in all_decoys_2
                proteins_dict[prot]['decoy1'] = proteins_dict[prot]['decoy'] and not proteins_dict[prot]['decoy2']
            proteins_dict[prot]['peptides set'].add(peptide)
            proteins_dict[prot]['TOP3'].append(ms1_i)
            proteins_dict[prot]['score'][peptide] = min(proteins_dict[prot]['score'].get(peptide, 1.0), pep)
            proteins_dict[prot]['PSMs'] += 1
            if pif > pif_threshold:
                for tag, val in zip(tagnames, tags):
                    proteins_dict[prot][tag] += val
        if not have_pif or pif > pif_threshold:
            for i, (tag, val) in enumerate(zip(tagnames, tags)):
                tagsums[i] += val

    if not have_pif:
        df1_f2.drop('PIF', axis='columns', inplace=True)
    df_proteins = pd.DataFrame.from_dict(proteins_dict, orient='index').reset_index()
    if path_to_fasta:
        df_proteins = process_fasta(df_proteins, path_to_fasta, decoy_prefix, decoy_infix)
    df_proteins['length'] = df_proteins['sequence'].apply(len)
    df_proteins['sq'] = df_proteins.apply(calc_sq, axis=1)
    df_proteins['peptides'] = df_proteins['peptides set'].apply(len)
    df_proteins['PSMs'] = df_proteins.apply(lambda x: max(x['PSMs'], x['peptides']), axis=1)
    # df_proteins.loc[:, ['PSMs', 'peptides']].max(axis=1)
    calc_NSAF(df_proteins)
    calc_TOP3(df_proteins)
    df_proteins['score'] = df_proteins['score'].apply(lambda x: np.prod(list(x.values())))
    norm = np.array(tagsums)
    df_proteins[tagnames] /= norm
    return df_proteins, norm


def calc_sq(df_raw):
    protein = df_raw['sequence']
    protein = protein.replace('L', 'I')
    peptides = df_raw['peptides set']
    if not protein:
        return 0
    psq = [False for x in protein]
    plen = len(protein)
    for pep in peptides:
        pep = pep.replace('L', 'I')
        csize = len(pep)
        for j in range(plen):
            if protein[j:j+csize] == pep:
                for y in range(csize):
                    psq[j + y] = True
    return float(sum(psq)) / len(psq) * 100


def get_output_basename(fname, suffix=''):
    basename = os.path.basename(fname)
    splt = os.path.splitext(basename)
    basename = splt[0]
    if 'pep' not in splt[1].lower():
        basename = os.path.splitext(basename)[0]
    return basename + suffix


def get_output_folder(folder, fname):
    if not folder:
        return os.path.dirname(os.path.realpath(fname))
    else:
        if not os.path.isdir(folder):
            os.mkdir(folder)
        return folder


def calc_RT(seq, RC):
    try:
        return achrom.calculate_RT(seq, RC)
    except Exception:
        return 0


def is_decoy(proteins, decoy_prefix, decoy_infix=False):
    if not decoy_infix:
        return all(z.startswith(decoy_prefix) for z in proteins)
    else:
        return all(decoy_infix in z for z in proteins)


def is_group_specific(proteins, group_prefix, group_infix, group_regex, decoy_prefix, decoy_infix=None):
    if group_regex:
        return all(re.search(group_regex, z) for z in proteins)
    if group_infix:
        return all(group_infix in z for z in proteins)
    if not decoy_infix:
        return all(z.startswith(decoy_prefix + group_prefix) or z.startswith(group_prefix) for z in proteins)
    return all(z.startswith(group_prefix) for z in proteins)


def is_decoy_2(proteins, decoy_set):
    return all(z in decoy_set for z in proteins)


def split_fasta_decoys(db, decoy_prefix, decoy_infix=None):
    decoy_dbnames = set()
    with fasta.read(db) as f:
        for protein in f:
            dbname = protein.description.split()[0]
            if (decoy_infix and decoy_infix in dbname) or dbname.startswith(decoy_prefix):
                decoy_dbnames.add(dbname)
    decoy_dbnames = sorted(decoy_dbnames)
    random.seed(SEED)
    all_decoys_2 = set(random.sample(decoy_dbnames, len(decoy_dbnames) // 2))
    logger.debug('Marking %s out of %s decoys as decoy2', len(all_decoys_2), len(decoy_dbnames))
    return all_decoys_2


def split_decoys(df):
    """Split decoys into decoy1 and decoy2 without FASTA file"""
    all_decoys = set()
    for proteins in df.loc[df.decoy, 'protein'].values:
        all_decoys.update(proteins)
    logger.debug('proteins: %s', proteins)
    all_decoys = sorted(all_decoys)  # sort is done for working of random SEED
    random.seed(SEED)
    all_decoys_2 = set(random.sample(all_decoys, int(len(all_decoys) / 2)))
    df['decoy2'] = df['protein'].apply(is_decoy_2, decoy_set=all_decoys_2)
    df['decoy1'] = df.apply(lambda x: x['decoy'] and not x['decoy2'], axis=1)
    return all_decoys_2


def remove_column_hit_rank(df):
    if 'hit_rank' in df.columns:
        return df[df['hit_rank'] == 1]
    else:
        return df


def parse_mods(df_raw):
    mods_counter = {}
    sequence, mods = df_raw['peptide'], df_raw['modifications']
    if isinstance(mods, list):
        for mod in mods:
            mod_mass, aa_ind = mod.split('@')
            mod_mass = float(mod_mass)
            aa_ind = int(aa_ind)
            if aa_ind == 0:
                aa = 'N_term'
                mod_mass = round(mod_mass - 1.007825, 3)
            elif aa_ind == len(sequence) + 1:
                aa = 'C_term'
                mod_mass = round(mod_mass - 17.002735, 3)
            else:
                aa = sequence[aa_ind - 1]
                mod_mass = round(mod_mass - mass.std_aa_mass[aa], 3)
            mod_name = 'mass shift %.3f at %s' % (mod_mass, aa)
            mods_counter[mod_name] = mods_counter.get(mod_name, 0) + 1
    return mods_counter


def parse_mods_msgf(df_raw):
    mods_counter = {}
    sequence, mods = df_raw['peptide'], df_raw['Modification']
    if isinstance(mods, list):
        for mod in mods:
            mod_mass, aa_ind = mod['monoisotopicMassDelta'], mod['location']
            if aa_ind == 0:
                aa = 'N_term'
                mod_mass = round(mod_mass, 3)
            elif aa_ind == len(sequence) + 1:
                aa = 'C_term'
                mod_mass = round(mod_mass, 3)
            else:
                aa = sequence[aa_ind - 1]
                mod_mass = round(mod_mass, 3)
            mod_name = 'mass shift %.3f at %s' % (mod_mass, aa)
            mods_counter[mod_name] = mods_counter.get(mod_name, 0) + 1
    return mods_counter


def add_mod_info(df_raw, mod):
    sequence, mods_counter = df_raw['peptide'], df_raw['mods_counter']
    mod_aa = mod.split(' at ')[1]
    if 'term' not in mod_aa and mod_aa not in sequence:
        return -1
    else:
        return (1 if mods_counter.get(mod, 0) >= 1 else 0)


def prepare_mods(df):
    all_mods = set()
    for cnt in df['mods_counter'].values:
        for k in cnt.keys():
            all_mods.add(k)
    for mod in all_mods:
        df[mod] = df.apply(add_mod_info, axis=1, mod=mod)


def prepare_dataframe(infile_path, decoy_prefix=None, decoy_infix=False, cleavage_rule=False, fdr=0.01, decoy2set=None):
    if not cleavage_rule:
        cleavage_rule = parser.expasy_rules['trypsin']
    if infile_path.lower().endswith('.pep.xml') or infile_path.lower().endswith('.pepxml'):
        df1 = pepxml.DataFrame(infile_path)
        ftype = 'pepxml'
    elif infile_path.lower().endswith('.mzid'):
        df1 = mzid.DataFrame(infile_path)
    else:
        raise WrongInputError()
    if not df1.shape[0]:
        raise EmptyFileError()

    if 'Morpheus Score' in df1.columns:
        df1 = df1[df1['Morpheus Score'] != 0]
        df1['expect'] = 1 / df1['Morpheus Score']
        df1['num_missed_cleavages'] = df1['peptide'].apply(lambda x: parser.num_sites(x, rule=cleavage_rule))

    if 'MS-GF:EValue' in df1.columns:
        # MSGF search engine
        ftype = 'msgf'
        df1['peptide'] = df1['PeptideSequence']
        df1['num_missed_cleavages'] = df1['peptide'].apply(lambda x: parser.num_sites(x, rule=cleavage_rule))
        df1['assumed_charge'] = df1['chargeState']
        df1['spectrum'] = df1['spectrumID']
        df1['massdiff'] = (df1['experimentalMassToCharge'] - df1['calculatedMassToCharge']) * df1['assumed_charge']
        df1['calc_neutral_pep_mass'] = df1['calculatedMassToCharge'] * df1['chargeState'] - df1['chargeState'] * 1.00727649
        df1['protein'] = df1['accession']
        df1['protein_descr'] = df1['protein description']
        df1['expect'] = df1['MS-GF:EValue']

    if set(df1['protein_descr'].str[0]) == {None}:
        # MSFragger
        logger.debug('Adapting MSFragger DataFrame.')
        logger.debug('Proteins before: %s', df1.loc[1, 'protein'])
        protein = df1['protein'].apply(lambda row: [x.split(None, 1) for x in row])
        df1['protein'] = protein.apply(lambda row: [x[0] for x in row])
        try:
            df1['protein_descr'] = protein.apply(lambda row: [x[1] for x in row])
        except IndexError:
            df1['protein_descr'] = protein.apply(lambda row: ['' for x in row])
        logger.debug('Proteins after: %s', df1.loc[1, 'protein'])

    # if any(None in set(df1['protein_descr'].str[0])):
    #     print('HERE')
    #     df1['protein_descr'] = df1.apply(lambda x: x['protein_descr'] if x['protein_descr'] else x['protein'], axis=1)
    df1.loc[pd.isna(df1['protein_descr']), 'protein_descr'] = df1.loc[pd.isna(df1['protein_descr']), 'protein']
    # try:
    #     df1['expect'] = 1.0 / df1['bions_score_neg'].values
    # except:
    #     pass

    df1 = df1[~pd.isna(df1['peptide'])]
    if 'MS1Intensity' not in df1:
        df1['MS1Intensity'] = 0.0
    df1['length'] = df1['peptide'].apply(len)
    df1 = df1[df1['length'] >= 6]
    df1['spectrum'] = df1['spectrum'].apply(lambda x: x.split(' RTINS')[0])
    if 'retention_time_sec' not in df1.columns:
        if 'scan start time' in df1.columns:
            df1['RT exp'] = df1['scan start time']
            df1 = df1.drop(['scan start time', ], axis=1)
        else:
            df1['RT exp'] = 0
    else:
        df1['RT exp'] = df1['retention_time_sec'] / 60
        df1 = df1.drop(['retention_time_sec', ], axis=1)

    df1['massdiff_int'] = df1['massdiff'].apply(lambda x: int(round(x, 0)))
    df1['massdiff_ppm'] = 1e6 * (df1['massdiff'] - df1['massdiff_int'] * 1.003354) / df1['calc_neutral_pep_mass']

    df1['decoy'] = df1['protein'].apply(is_decoy, decoy_prefix=decoy_prefix, decoy_infix=decoy_infix)
    if not df1.decoy.sum():
        raise NoDecoyError()
    if decoy2set is None:
        decoy2set = split_decoys(df1)
    else:
        df1['decoy2'] = df1['protein'].apply(lambda p: all(x in decoy2set for x in p))
        df1['decoy1'] = df1['decoy'] & (~df1['decoy2'])
    df1 = remove_column_hit_rank(df1)

    if ftype == 'pepxml':
        df1['mods_counter'] = df1.apply(parse_mods, axis=1)
    elif ftype == 'msgf':
        df1['mods_counter'] = df1.apply(parse_mods_msgf, axis=1)
    prepare_mods(df1)

    pep_ratio = df1['decoy2'].sum() / df1['decoy'].sum()
    df1_f = filter_custom(df1[~df1['decoy1']], fdr=fdr, key='expect', is_decoy='decoy2',
        reverse=False, remove_decoy=False, ratio=pep_ratio, formula=1, correction=None, loglabel='PSMs default')
    num_psms_def = df1_f[~df1_f['decoy2']].shape[0]
    logger.info('Default target-decoy filtering, 1%% PSM FDR: Number of target PSMs = %d', num_psms_def)
    try:
        logger.info('Calibrating retention model...')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            retention_coefficients = achrom.get_RCs_vary_lcp(df1_f['peptide'].values, df1_f['RT exp'].values)
        df1_f['RT pred'] = df1_f['peptide'].apply(lambda x: calc_RT(x, retention_coefficients))
        df1['RT pred'] = df1['peptide'].apply(lambda x: calc_RT(x, retention_coefficients))
        _, _, r_value, std_value = aux.linear_regression(df1_f['RT pred'], df1_f['RT exp'])
        logger.info('RT model training results: R^2 = %f , std = %f', r_value**2, std_value)
        df1['RT diff'] = df1['RT pred'] - df1['RT exp']
        logger.info('Retention model calibrated successfully.')
    except Exception:
        logger.warning('Retention times are probably missing in input file.')
        df1['RT pred'] = df1['peptide'].apply(lambda x: calc_RT(x, achrom.RCs_krokhin_100A_tfa))
        df1['RT diff'] = df1['RT exp']
    return df1, decoy2set


_standard_features = {'calc_neutral_pep_mass', 'bscore', 'yscore',
                'massdiff', 'massdiff_ppm', 'nextscore', 'RT pred', 'RT diff',
                'sumI', 'RT exp', 'precursor_neutral_mass', 'massdiff_int',
                'num_missed_cleavages', 'tot_num_ions', 'num_matched_ions', 'length',
                'ScoreRatio', 'Energy', 'MS2IonCurrent', 'MeanErrorTop7', 'sqMeanErrorTop7', 'StdevErrorTop7',
                'MS-GF:DeNovoScore', 'MS-GF:EValue', 'MS-GF:RawScore', 'MeanErrorAll',
                'MeanRelErrorAll', 'MeanRelErrorTop7', 'NumMatchedMainIons', 'StdevErrorAll',
                'StdevErrorTop7', 'StdevRelErrorAll', 'StdevRelErrorTop7', 'NTermIonCurrentRatio',
                'CTermIonCurrentRatio', 'ExplainedIonCurrentRatio', 'fragmentMT', 'ISOWIDTHDIFF',
                'MS1Intensity', 'sumI_to_MS1Intensity', 'nextscore_std', 'IPGF', 'IPGF2', 'hyperscore', 'PIF'}


def get_features(dataframe):
    feature_columns = dataframe.columns
    columns_to_remove = []
    for feature in feature_columns:
        if feature not in _standard_features:
            if not feature.startswith('mass shift') and not feature.startswith('matched_y') and not feature.startswith('matched_b'):
                columns_to_remove.append(feature)
    feature_columns = feature_columns.drop(columns_to_remove)
    return sorted(feature_columns)


def get_X_array(df, feature_columns):
    return df.loc[:, feature_columns].values


def get_Y_array(df):
    return df.loc[:, 'decoy1'].values.astype(float)


def variant_peptides(allowed_peptides, group_prefix, group_infix, group_regex):
    if allowed_peptides:
        with open(allowed_peptides) as f:
            allowed_peptides = set(pseq.strip().split()[0] for pseq in f)
    else:
        allowed_peptides = None

    return allowed_peptides, group_prefix, group_infix, group_regex


def filename(outfolder, outbasename, ftype):
    type_suffix = {
        'psm_full': '_PSMs_full.tsv',
        'psm': '_PSMs.tsv',
        'pepxml': '.scavager.pep.xml',
        'peptide': '_peptides.tsv',
        'protein': '_proteins.tsv',
        'protein_group': '_protein_groups.tsv'
    }
    return os.path.join(outfolder, outbasename + type_suffix[ftype])


def get_cat_model(df, feature_columns):
    logger.info('Starting machine learning...')

    # logger.info(df.shape)
    # df['orig_spectrum'] = df['spectrum'].apply(lambda x: x.split('.')[-3])
    # df['massdiff_abs'] = df['massdiff'].abs()
    # df = df.sort_values(by='massdiff_abs')
    # df = df.drop_duplicates(subset = ['orig_spectrum', 'peptide'])
    # logger.info(df.shape)

    train, test = train_test_split(df, test_size=0.3, random_state=SEED)
    x_train = get_X_array(train, feature_columns)
    y_train = get_Y_array(train)
    x_test = get_X_array(test, feature_columns)
    y_test = get_Y_array(test)
    np.random.seed(SEED)
    # model = CatBoostClassifier(iterations=1000, learning_rate=0.05, depth=10, loss_function='Logloss', logging_level='Silent', random_seed=SEED)
    # model.fit(x_train, y_train, use_best_model=True, eval_set=(x_test, y_test))

    model = CatBoostClassifier(iterations=10000, learning_rate=0.01, depth=8, loss_function='Logloss', eval_metric='Logloss',
                               od_type='Iter', od_wait=33, random_state=SEED, logging_level='Silent')
    model.fit(x_train, y_train, use_best_model=True, eval_set=(x_test, y_test))
    best_iter = model.best_iteration_
    logger.debug('Best iteration: %d', best_iter)
    ln_rt = max(0.001, round(0.01 * best_iter / 1000, 3))
    model = CatBoostClassifier(iterations=10000, learning_rate=ln_rt, depth=8, loss_function='Logloss', eval_metric='Logloss',
                               od_type='Iter', od_wait=33, random_state=SEED, logging_level='Silent')
    model.fit(x_train, y_train, use_best_model=True, eval_set=(x_test, y_test))
    best_iter = model.best_iteration_
    logger.debug('Best iteration: %d', best_iter)
    X = get_X_array(df, feature_columns)
    y = get_Y_array(df)
    model = CatBoostClassifier(iterations=int(best_iter / 0.7), learning_rate=ln_rt, depth=8, loss_function='Logloss',
                               random_state=SEED, logging_level='Silent')
    model.fit(X, y)
    logger.debug('Feature importance:')
    for fi, fn in sorted(zip(model.feature_importances_, feature_columns), key=lambda x: x[0])[::-1]:
        logger.debug('%s: %s', fi, fn)

    logger.info('Machine learning is finished.')

    return model


def calc_PEP(df, pep_ratio=1.0, reduced=False):
    if not reduced:
        feature_columns = get_features(df)
        cat_model = get_cat_model(df, feature_columns)
        x_all = get_X_array(df, feature_columns)
        df['ML score'] = cat_model.predict_proba(x_all)[:, 1]
    else:
        df['ML score'] = df['expect']

    df0_t = df[~df['decoy']]
    df0_d = df[df['decoy']]
    df0_d = df0_d[~df0_d['decoy1']]

    binsize = min(get_fdbinsize(df0_t['ML score'].values), get_fdbinsize(df0_d['ML score'].values))
    tmp = np.concatenate([df0_t['ML score'].values, df0_d['ML score'].values])
    minv = df['ML score'].min()
    maxv = df['ML score'].max()
    lbin_s = scoreatpercentile(tmp, 1.0)
    lbin = minv
    if lbin_s and abs((lbin - lbin_s) / lbin_s) > 1.0:
        lbin = lbin_s * 1.05
    rbin_s = scoreatpercentile(tmp, 99.0)
    rbin = maxv
    if rbin_s and abs((rbin - rbin_s) / rbin_s) > 1.0:
        rbin = rbin_s * 1.05
    rbin += 1.5 * binsize
    logger.debug('cbins: lbin = %s, rbin = %s, binsize = %s', lbin, rbin, binsize)
    cbins = np.arange(lbin, rbin + 2 * binsize, binsize)

    H1, b1 = np.histogram(df0_d['ML score'].values, bins=cbins)
    H2, b2 = np.histogram(df0_t['ML score'].values, bins=cbins)

    H2[H2 == 0] = 1
    H1_2 = H1 * (1 + 1. / pep_ratio) / H2
    ir = IsotonicRegression(y_min=0, y_max=1.0)
    ir.fit(b1[:-1], H1_2)
    df['PEP'] = ir.predict(df['ML score'].values)

    pep_min = df['ML score'].min()
    df['log_score'] = np.log10(df['ML score'] - ((pep_min - 1e-15) if pep_min < 0 else 0))


def calc_qvals(df, ratio):
    logger.debug('Q-value calculation started...')
    df_t_1 = aux.qvalues(df[~df['decoy1']], key='ML score', is_decoy='decoy2',
        remove_decoy=False, formula=1, full_output=True, ratio=ratio, correction=1)
    df_t = aux.qvalues(df[~df['decoy1']], key='ML score', is_decoy='decoy2',
        remove_decoy=False, formula=1, full_output=True, ratio=ratio, correction=0)
    df.loc[~df['decoy1'], 'q'] = df_t_1['q']
    df.loc[~df['decoy1'], 'q_uncorrected'] = df_t['q']
    df.loc[df['decoy1'], 'q'] = None
    df.loc[df['decoy1'], 'q_uncorrected'] = None


_columns_to_output = {
    'psm_full': ['peptide', 'length', 'spectrum', 'q', 'q_uncorrected', 'ML score', 'modifications', 'assumed_charge',
         'num_missed_cleavages', 'num_tol_term', 'peptide_next_aa',
         'peptide_prev_aa', 'calc_neutral_pep_mass', 'massdiff_ppm', 'massdiff_int', 'RT exp', 'RT pred',
         'RT diff', 'protein', 'protein_descr', 'decoy', 'decoy1', 'decoy2', 'PEP',
         'MS1Intensity', 'ISOWIDTHDIFF'],
    'psm': ['peptide', 'length', 'spectrum', 'q', 'q_uncorrected','ML score', 'modifications', 'assumed_charge',
         'num_missed_cleavages', 'num_tol_term', 'peptide_next_aa',
         'peptide_prev_aa', 'calc_neutral_pep_mass', 'massdiff_ppm', 'massdiff_int', 'RT exp', 'RT pred',
         'protein', 'protein_descr', 'decoy', 'PEP', 'MS1Intensity', 'ISOWIDTHDIFF', 'PIF'],
    'peptide': ['peptide', '#PSMs', 'length', 'spectrum', 'q', 'q_uncorrected', 'ML score', 'modifications',
         'assumed_charge', 'num_missed_cleavages', 'num_tol_term', 'peptide_next_aa',
         'peptide_prev_aa', 'calc_neutral_pep_mass', 'massdiff_ppm', 'massdiff_int', 'RT exp',
         'RT pred', 'protein', 'protein_descr', 'decoy', 'PEP', 'MS1Intensity', 'ISOWIDTHDIFF', 'PIF'],
    'protein': ['dbname', 'description', 'PSMs', 'peptides', 'NSAF', 'TOP3', 'sq', 'score', 'q', 'q_uncorrected',
         'length', 'all proteins', 'groupleader'],
    }


def get_columns_to_output(columns, out_type):
    present = set(columns)
    order = _columns_to_output[out_type]
    labels = [label for label in order if label in present]
    if out_type == 'psm_full':
        labels.extend(present.difference(order))
    if out_type != 'psm_full':
        for c in columns:
            if c[:4] == 'tag_':
                labels.append(c)
    logger.debug('Writing out %s table, q_uncorrected present: %s', out_type, 'q_uncorrected' in labels)
    return labels


def calc_psms(df, df_filt):
    peptides = Counter(df_filt['peptide'])
    df['#PSMs'] = df['peptide'].apply(lambda x: peptides.get(x, 0))



from __future__ import division
from scipy.stats import scoreatpercentile
from pyteomics import mass
from collections import Counter
import os.path
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
try:
    import seaborn
    seaborn.set(rc={'axes.facecolor':'#ffffff'})
    seaborn.set_style('whitegrid')
except ImportError:
    pass
import re
import sys
import logging
logger = logging.getLogger(__name__)
logging.getLogger('matplotlib.font_manager').disabled = True
redcolor = '#FC6264'
bluecolor = '#70aed1'
greencolor = '#8AA413'
aa_color_1 = '#fca110'
aa_color_2 = '#a41389'

def _get_sf(fig):
    return isinstance(fig, str) if sys.version_info.major == 3 else isinstance(fig, basestring)


def get_basic_distributions(df):
    mz_array = ((df['calc_neutral_pep_mass'] + df['assumed_charge'] * 1.007276 ) / df['assumed_charge']).values
    rt_exp_array = (df['RT exp']).values
    lengths_array = df['length'].values
    return mz_array, rt_exp_array, lengths_array


def get_descriptor_array(df, df_f, dname):
    array_t = df[~df['decoy']][dname].values
    array_d = df[df['decoy']][dname].values
    array_v = df_f[dname].values
    return array_t, array_d, array_v


def plot_hist_basic(array_all, array_valid, fig, subplot_max_x, subplot_i,
        xlabel, ylabel='# of identifications', bin_size_one=False):
    separate_figures = _get_sf(fig)
    if separate_figures:
        plt.figure()
    else:
        fig.add_subplot(subplot_max_x, 3, subplot_i)
    cbins = get_bins((array_all, array_valid), bin_size_one)
    plt.hist(array_all, bins=cbins, color=redcolor, alpha=0.8, edgecolor='#EEEEEE')
    plt.hist(array_valid, bins=cbins, color=greencolor, alpha=0.8, edgecolor='#EEEEEE')
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    if separate_figures:
        plt.savefig(outpath(fig, xlabel, '.png'))
        plt.close()


def plot_basic_figures(df, df_f, fig, subplot_max_x, subplot_start, idtype):
    logger.debug('Plotting %s figures...', idtype)
    logger.debug('Getting basic distributions (all)')
    mz_array, rt_exp_array, lengths_array = get_basic_distributions(df)
    logger.debug('Getting basic distributions (filtered)')
    mz_array_valid, rt_exp_array_valid, lengths_array_valid = get_basic_distributions(df_f)

    logger.debug('Plotting %s m/z distributions.', idtype)
    plot_hist_basic(mz_array, mz_array_valid, fig, subplot_max_x, subplot_i=subplot_start,
                     xlabel='%s, precursor m/z' % (idtype, ))
    subplot_start += 1
    logger.debug('Plotting %s RT distributions.', idtype)
    plot_hist_basic(rt_exp_array, rt_exp_array_valid, fig, subplot_max_x, subplot_i=subplot_start,
                     xlabel='%s, RT experimental' % (idtype, ))
    subplot_start += 1
    logger.debug('Plotting %s length distributions.', idtype)
    plot_hist_basic(lengths_array, lengths_array_valid, fig, subplot_max_x, subplot_i=subplot_start,
                     xlabel='%s, peptide length' % (idtype, ), bin_size_one=True)


def plot_protein_figures(df, df_f, fig, subplot_max_x, subplot_start):
    logger.debug('Plotting protein figures: NSAF')
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='LOG10_NSAF'), fig, subplot_max_x, subplot_start, xlabel='LOG10(NSAF)')
    logger.debug('Plotting protein figures: SQ')
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='sq'), fig, subplot_max_x, subplot_start+1, xlabel='sequence coverage')


def plot_hist_descriptor(inarrays, fig, subplot_max_x, subplot_i, xlabel, ylabel='# of identifications'):
    logger.debug('Plotting descriptor histogram: %s', xlabel)
    separate_figures = _get_sf(fig)
    if separate_figures:
        plt.figure()
    else:
        fig.add_subplot(subplot_max_x, 3, subplot_i)
    array_t, array_d, array_v = inarrays
    cbins, width = get_bins_for_descriptors(inarrays)
    H1, _ = np.histogram(array_d, bins=cbins)
    H2, _ = np.histogram(array_t, bins=cbins)
    H3, _ = np.histogram(array_v, bins=cbins)
    plt.bar(cbins[:-1], H1, width, align='center',color=redcolor, alpha=0.4, edgecolor='#EEEEEE')
    plt.bar(cbins[:-1], H2, width, align='center',color=bluecolor, alpha=0.4, edgecolor='#EEEEEE')
    plt.bar(cbins[:-1], H3, width, align='center',color=greencolor, alpha=1, edgecolor='#EEEEEE')
    cbins = np.append(cbins[0], cbins)
    H1 = np.append(np.append(0, H1), 0)
    H2 = np.append(np.append(0, H2), 0)
    cbins -= width / 2
    plt.step(cbins, H2, where='post', color=bluecolor, alpha=0.8)
    plt.step(cbins, H1, where='post', color=redcolor, alpha=0.8)
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    if 'mass shift' in xlabel:
        plt.xlim(-1.5, 1.5)
        plt.xticks([-1, 0, 1], ['NA', 'unmodified', 'modified'])
    elif width == 1.0:
        plt.xticks(np.arange(int(cbins[0]), cbins[-1], 1))
        plt.gcf().canvas.draw()
    if separate_figures:
        plt.savefig(outpath(fig, xlabel, '.png'))
        plt.close()


def plot_legend(fig, subplot_max_x, subplot_start):
    ax = fig.add_subplot(subplot_max_x, 3, subplot_start)
    legend_elements = [Patch(facecolor=greencolor, label='Positive IDs'),
                    Patch(facecolor=bluecolor, label='Targets'),
                    Patch(facecolor=redcolor, label='Decoys')]
    ax.legend(handles=legend_elements, loc='center', prop={'size': 24})
    ax.set_axis_off()


def plot_aa_stats(df_f, df_proteins_f, fig, subplot_max_x, subplot_i):
    separate_figures = _get_sf(fig)
    if separate_figures:
        plt.figure()
    else:
        fig.add_subplot(subplot_max_x, 3, subplot_i)

    # Generate list of 20 standart amino acids
    std_aa_list = list(mass.std_aa_mass.keys())
    std_aa_list.remove('O')
    std_aa_list.remove('U')

    # Count identified amino acids
    aa_exp = Counter()
    for pep in set(df_f['peptide']):
        for aa in pep:
            aa_exp[aa] += 1

    # Count theoretical amino acids
    aa_theor = Counter()
    for prot_seq in set(df_proteins_f['sequence'].values):
        for aa in prot_seq:
            aa_theor[aa] += 1

    aa_exp_sum = sum(aa_exp.values())
    aa_theor_sum = sum(aa_theor.values())
    lbls, vals = [], []
    for aa in sorted(std_aa_list):
        if aa_theor.get(aa, 0):
            lbls.append(aa)
            vals.append((aa_exp.get(aa, 0)/aa_exp_sum) / (aa_theor[aa]/aa_theor_sum))
    # std_val = np.std(vals)
    clrs = [aa_color_1 if abs(x-1) < 0.4 else aa_color_2 for x in vals]
    plt.bar(range(len(vals)), vals, color=clrs)
    plt.xticks(range(len(lbls)), lbls)
    plt.hlines(1.0, range(len(vals))[0]-1, range(len(vals))[-1]+1)
    plt.ylabel('amino acid ID rate')


def calc_max_x_value(df, df_proteins):
    cnt = 7 # number of basic figures
    peptide_columns = set(df.columns)
    features_list = ['massdiff_ppm', 'RT diff', 'fragmentMT', 'num_missed_cleavages', 'assumed_charge', 'log_score', 'ISOWIDTHDIFF', 'MS1Intensity']
    for feature in features_list:
        if feature in peptide_columns:
            cnt += 1
    for feature in peptide_columns:
        if feature.startswith('mass shift'):
            cnt += 1
    if len(set(df['massdiff_int'])) > 1:
        cnt += 1
    if 'LOG10_NSAF' in df_proteins.columns:
        cnt += 3 # add for NSAF, sequence coverage and aa_stats
    return cnt // 3 + (1 if (cnt % 3) else 0)


def plot_descriptors_figures(df, df_f, fig, subplot_max_x, subplot_start):
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='massdiff_ppm'), fig, subplot_max_x, subplot_start, xlabel='precursor mass difference, ppm')
    subplot_start += 1
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='RT diff'), fig, subplot_max_x, subplot_start, xlabel='RT difference, min')
    subplot_start += 1
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='num_missed_cleavages'), fig, subplot_max_x, subplot_start, xlabel='missed cleavages')
    subplot_start += 1
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='assumed_charge'), fig, subplot_max_x, subplot_start, xlabel='precursor charge')
    subplot_start += 1
    if 'fragmentMT' in df.columns:
        plot_hist_descriptor(get_descriptor_array(df, df_f, dname='fragmentMT'), fig, subplot_max_x, subplot_start, xlabel='median fragment error, Da')
        subplot_start += 1
    if 'ISOWIDTHDIFF' in df.columns:
        plot_hist_descriptor(get_descriptor_array(df, df_f, dname='ISOWIDTHDIFF'), fig, subplot_max_x, subplot_start, xlabel='isolation mass error, Da')
        subplot_start += 1
    if 'MS1Intensity' in df.columns:
        plot_hist_descriptor(get_descriptor_array(df, df_f, dname='MS1Intensity'), fig, subplot_max_x, subplot_start, xlabel='MS1 Intensity')
        subplot_start += 1

    if len(set(df['massdiff_int'])) > 1:
        plot_hist_descriptor(get_descriptor_array(df, df_f, dname='massdiff_int'), fig, subplot_max_x, subplot_start, xlabel='isotope mass difference, Da')
        subplot_start += 1
    for df_col in df.columns:
        if df_col.startswith('mass shift'):
            plot_hist_descriptor(get_descriptor_array(df, df_f, dname=df_col), fig, subplot_max_x, subplot_start, xlabel=df_col)
            subplot_start += 1
    plot_hist_descriptor(get_descriptor_array(df, df_f, dname='log_score'), fig, subplot_max_x, subplot_start, xlabel='LOG10(ML score)')
    subplot_start += 1
    separate_figures = _get_sf(fig)
    if not separate_figures:
        plot_legend(fig, subplot_max_x, subplot_start)
    subplot_start += 1


def get_bins(inarrays, bin_size_one=False):
    tmp = np.concatenate(inarrays)
    minv = tmp.min()
    maxv = tmp.max()
    if bin_size_one:
        return np.arange(minv, maxv+1, 1)
    else:
        return np.linspace(minv, maxv+1, num=100)


def get_bins_for_descriptors(inarrays):
    tmp = np.concatenate(inarrays)
    minv = tmp.min()
    maxv = tmp.max()
    if len(set(tmp)) <= 15:
        return np.arange(minv, maxv + 2, 1.0), 1.0
    binsize = False
    for inar in inarrays:
        binsize_tmp = get_fdbinsize(inar)
        if not binsize or binsize > binsize_tmp:
            binsize = binsize_tmp
    # binsize = get_fdbinsize(tmp)
    if binsize < float(maxv - minv) / 300:
        binsize = float(maxv - minv) / 300

    lbin_s = scoreatpercentile(tmp, 1.0)
    lbin = minv
    if lbin_s and abs((lbin - lbin_s) / lbin_s) > 1.0:
        lbin = lbin_s * 1.05
    rbin_s = scoreatpercentile(tmp, 99.0)
    rbin = maxv
    if rbin_s and abs((rbin - rbin_s) / rbin_s) > 1.0:
        rbin = rbin_s * 1.05
    rbin += 1.5 * binsize
    logger.debug('get_bins_for_descriptors: lbin = %s, rbin = %s, binsize = %s', lbin, rbin, binsize)
    return np.arange(lbin, rbin + binsize, binsize), binsize


def get_fdbinsize(data_list):
    """Calculates the Freedman-Diaconis bin size for
    a data set for use in making a histogram
    Arguments:
    data_list:  1D Data set
    Returns:
    optimal_bin_size:  F-D bin size
    """
    if not isinstance(data_list, np.ndarray):
        data_list = np.array(data_list)
    isnan = np.isnan(data_list)
    logger.debug('Number of nans: %s', isnan.sum())
    data_list = np.sort(data_list[~isnan])
    upperquartile = scoreatpercentile(data_list, 75)
    lowerquartile = scoreatpercentile(data_list, 25)
    iqr = upperquartile - lowerquartile
    logger.debug('IQR: %s, data size: %s', iqr, len(data_list))
    optimal_bin_size = 2. * iqr / len(data_list) ** (1. / 3.)
    logger.debug('Calculated optimal bin size: %s.', optimal_bin_size)
    MINBIN = 1e-8
    if optimal_bin_size < MINBIN:
        logger.debug('Increasing bin size to %s.', MINBIN)
        return MINBIN
    return optimal_bin_size


def normalize_fname(s):
    return re.sub(r'[<>:\|/?*]', '', s)


def outpath(outfolder, s, ext='.png'):
    return os.path.join(outfolder, normalize_fname(s) + ext)


def plot_outfigures(df, df_f, df_peptides, df_peptides_f, outfolder, outbasename, df_proteins, df_proteins_f, separate_figures=False):
    if not separate_figures:
        fig = plt.figure(figsize=(16, 12))
        dpi = fig.get_dpi()
        fig.set_size_inches(3000.0/dpi, 3000.0/dpi)
    else:
        outfolder = os.path.join(outfolder, outbasename + '_figures')
        if not os.path.isdir(outfolder):
            os.makedirs(outfolder)
        fig = outfolder
    subplot_max_x = calc_max_x_value(df, df_proteins)
    descriptor_start_index = 7
    logger.debug('Plotting PSM figures...')
    plot_basic_figures(df, df_f, fig, subplot_max_x, 1, 'PSMs')
    logger.debug('Plotting peptide figures...')
    plot_basic_figures(df_peptides, df_peptides_f, fig, subplot_max_x, 4, 'peptides')
    if 'LOG10_NSAF' in df_proteins.columns:
        logger.debug('Plotting protein figures...')
        try:
            plot_protein_figures(df_proteins, df_proteins_f, fig, subplot_max_x, 7)
            logger.debug('Plotting AA stats figures...')
            plot_aa_stats(df_f, df_proteins_f, fig, subplot_max_x, 9)
            descriptor_start_index += 3
        except ZeroDivisionError:
            logger.warning('There was an error in protein stats calculation. No protein figures will be plotted.')
    logger.debug('Plotting descriptor figures...')
    plot_descriptors_figures(df, df_f, fig, subplot_max_x, descriptor_start_index)
    plt.grid(color='#EEEEEE')
    plt.tight_layout()
    if not separate_figures:
        plt.savefig(os.path.join(outfolder, outbasename) + '.png')
        plt.close()
    logger.info('Figures saved.')



#!/usr/bin/env python

'''
setup.py file for Scavager
'''
from setuptools import setup
version = open('VERSION').readline().strip()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name             = 'Scavager',
    version          = version,
    description      = '''Proteomics post-search algorithm''',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    author           = 'Mark Ivanov & Lev Levitsky & Julia Bubis',
    author_email     = 'pyteomics@googlegroups.com',
    url              = 'https://github.com/markmipt/scavager',
    packages         = ['scavager', ],
    install_requires = [line.strip() for line in open('requirements.txt')],
    classifiers      = ['Intended Audience :: Science/Research',
                        'Programming Language :: Python :: 3',
                        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    license          = 'License :: OSI Approved :: Apache Software License',
    entry_points     = {'console_scripts': ['scavager = scavager.search:run',
                                            'scav2diffacto = scavager.scav2diffacto:run',
                                            'scav2nsaf = scavager.scav2nsaf:run']}
    )



from __future__ import division
import os.path
import logging
import ast

import pandas as pd
from pyteomics import auxiliary as aux
from catboost import CatBoostError
try:
    from pyteomics import pepxmltk
except ImportError:
    pepxmltk = None
from . import utils
from .utils_figures import plot_outfigures
logger = logging.getLogger(__name__)

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def process_files(args):
    """Run Scavager for multiple files (`args['file']` should be a list of file names)
    and possibly for their union.

    Parameters
    ----------

    args : dict
        A dictionary of parameters as produced from argparse in :py:func:`search.run`.

    Returns
    -------
    out : int
        Exit code. 0 for success, 1 for empty (union) result, negative for errors
        (first encountered error code is returned).
    """
    files = args['file']
    N = len(files)
    logger.info('%d file(s) to process.', N)
    cargs = args.copy()
    if args['union'] or args['quick_union']:
        if not args['database']:
            logger.error('--database is required for union calculation.')
            return -101
    if args['create_pepxml'] and pepxmltk is None:
        logger.error('pepxmltk is required for --create-pepxml. Please install it.')
        return -102
    if args['database']:
        decoy_prots_2 = utils.split_fasta_decoys(args['database'], args['prefix'], args['infix'])
    else:
        decoy_prots_2 = None
        logger.info('Database file not provided. Decoy randomization will be done per PSM file.')
    errors = 0
    retvalues = []
    if not args['quick_union']:
        for f in files:
            cargs['file'] = f
            retv = process_file(cargs, decoy2=decoy_prots_2)
            retvalues.append(retv)
            if -10 < retv < 0:
                logger.info('Stopping due to previous errors.')
                return retv
            if retv < 0:
                errors += 1
    else:
        logger.info('Skipping individual file processing.')
    if N == 1:
        return retv
    if errors >= N - 1:
        if N > 1:
            logger.info('Union will not be run because %s out of %s files were processed with errors.', errors, N)
        return retvalues
    if args['union'] or args['quick_union']:
        logger.info('Starting the union calculation...')
        psm_full_dfs = []
        for file in files:
            outfolder = utils.get_output_folder(args['output'], file)
            outbasename = utils.get_output_basename(file, args['name_suffix'])
            csvname = utils.filename(outfolder, outbasename, 'psm_full')
            try:
                df = pd.read_csv(csvname, sep='\t', converters={
                    key: ast.literal_eval for key in [
                        'protein', 'peptide_next_aa', 'peptide_prev_aa', 'num_tol_term',
                        'protein_descr', 'modifications', 'mods_counter']})
                df['file'] = os.path.basename(file)
                psm_full_dfs.append(df)
            except FileNotFoundError as e:
                logger.debug('Exception when reading %s: %s (%s)', file, e, e.args)
                logger.warning('File %s not found, skipping...', csvname)
        if not psm_full_dfs:
            logger.error('No PSMs found for union calculation.')
            retvalues.append(1)
            return retvalues
        all_psms = pd.concat(psm_full_dfs, sort=False)
        all_psms.reset_index(inplace=True, drop=True)
        logger.debug('Recovered PSMs for analysis: %s, of those: %s decoy1, %s decoy2, %s have q < %s',
            all_psms.shape, all_psms.decoy1.sum(), all_psms.decoy2.sum(),
            (all_psms['q'] < args['fdr'] / 100).sum(), args['fdr'] / 100)
        utils.prepare_mods(all_psms)
        q_label = 'q'
        if not (args['no_correction'] or args['force_correction']):
            logger.info('Using the corrected q-values for union.')
            all_psms_f2 = all_psms[(~all_psms['decoy1']) & (all_psms[q_label] < args['fdr'] / 100)]
            if not all_psms_f2.shape[0]:
                q_label = 'q_uncorrected'
                logger.info('No union results with correction. Disabling...')
        if args['no_correction']:
            q_label = 'q_uncorrected'
        logger.debug('Filtering union PSMs by %s.', q_label)
        all_psms_f2 = all_psms.loc[(~all_psms['decoy1']) & (all_psms[q_label] < args['fdr'] / 100)].copy()

        peptides, peptides_f, proteins, proteins_f, protein_groups = build_output_tables(all_psms,
            all_psms_f2, decoy_prots_2, args, 'PEP', calc_qvals=False)
        if peptides is None:
            logger.warning('No peptides identified in union.')
            retvalues.append(1)
            return retvalues

        logger.debug('Protein FDR in full table: %f%%', 100 * aux.fdr(proteins, is_decoy='decoy2'))

        write_tables(outfolder, 'union' + args['name_suffix'] + args['union_name_suffix'],
            all_psms, all_psms_f2, peptides_f, proteins_f, protein_groups)
        if args['create_pepxml']:
            pepxmltk.easy_write_pepxml(files, utils.filename(outfolder, 'union', 'pepxml'),
            set(all_psms_f2.loc[~all_psms_f2['decoy2'], 'spectrum']))

        if len(all_psms_f2[~all_psms_f2['decoy2']]) >= 3:
            plot_outfigures(all_psms, all_psms_f2[~all_psms_f2['decoy2']], peptides,
                peptides_f[~peptides_f['decoy2']],
                outfolder, 'union' + args['name_suffix'] + args['union_name_suffix'], df_proteins=proteins,
                df_proteins_f=proteins_f[~proteins_f['decoy2']],
                separate_figures=args['separate_figures'])

        logger.info('Union calculation complete.')
        retvalues.append(0)
    return retvalues


def filter_dataframe(df1, outfdr, correction, allowed_peptides, group_prefix, group_infix, group_regex, decoy_prefix, decoy_infix):
    d2 = df1['decoy2'].sum()
    d = df1['decoy'].sum()
    pep_ratio = d2 / d
    logger.debug('Peptide ratio for ML: %s (%s / %s)', pep_ratio, d2, d)
    df1_f = utils.filter_custom(df1[~df1['decoy1']], fdr=outfdr, key='expect', is_decoy='decoy2',
        reverse=False, remove_decoy=False, ratio=pep_ratio, formula=1, correction=correction, loglabel='PSMs default')
    num_psms_def = df1_f[~df1_f['decoy2']].shape[0]
    if not num_psms_def:
        logger.warning('No valid PSMs at all are found at %s%% FDR. Aborting the analysis.', 100 * outfdr)
        return df1, None
    if num_psms_def < 100:
        logger.warning('Not enough statistics for ML training (%d PSMs is less than 100).', num_psms_def)
        utils.calc_PEP(df1, pep_ratio=pep_ratio, reduced=True)
    else:
        utils.calc_PEP(df1, pep_ratio=pep_ratio)
        df1_f2 = utils.filter_custom(df1[~df1['decoy1']], fdr=outfdr, key='ML score',
            is_decoy='decoy2', reverse=False, remove_decoy=False, ratio=pep_ratio, formula=1, correction=correction, loglabel='PSMs')
        if df1_f2[~df1_f2['decoy2']].shape[0] < num_psms_def:
            logger.warning('Machine learning works worse than default filtering: %d vs %d PSMs.', df1_f2.shape[0], num_psms_def)
            logger.warning('Using only default search scores for machine learning...')
            utils.calc_PEP(df1, pep_ratio=pep_ratio, reduced=True)

    if allowed_peptides or group_prefix or group_infix or group_regex:
        prev_num = df1.shape[0]
        if allowed_peptides:
            df1 = df1[df1['peptide'].apply(lambda x: x in allowed_peptides)]
        else:
            logger.debug('Protein column looks like this: %s', df1['protein'].iloc[0])
            df1 = df1[df1['protein'].apply(utils.is_group_specific,
                group_prefix=group_prefix, group_infix=group_infix, group_regex=group_regex,
                decoy_prefix=decoy_prefix, decoy_infix=decoy_infix)]

        logger.info('%.1f%% of identifications were dropped during group-specific filtering.',
            (100 * float(prev_num - df1.shape[0]) / prev_num))

        if df1[df1['decoy']].shape[0] == 0:
            logger.warning('0 decoy identifications are present in the group. Please check'
            'that allowed_peptides contains decoy peptides or that decoy proteins have group_prefix/infix!')

    pep_ratio = df1['decoy2'].sum() / df1['decoy'].sum()
    logger.debug('Peptide ratio within group: %s', pep_ratio)
    df1_f2 = utils.filter_custom(df1[~df1['decoy1']], fdr=outfdr, key='ML score', is_decoy='decoy2',
        reverse=False, remove_decoy=False, ratio=pep_ratio, formula=1, correction=correction, loglabel='PSMs in group')

    return df1, df1_f2


def build_output_tables(df1, df1_f2, decoy2, args, key='ML score', calc_qvals=True):
    if args['database']:
        path_to_fasta = os.path.abspath(args['database'])
    else:
        path_to_fasta = None
    outfdr = args['fdr'] / 100
    correction = False if args['no_correction'] else (args['force_correction'] or None)
    pep_ratio = df1['decoy2'].sum() / df1['decoy'].sum()
    logger.debug('Peptide ratio for q-value calculation: %s', pep_ratio)

    if calc_qvals:
        utils.calc_qvals(df1, ratio=pep_ratio)

    if df1_f2.shape[0]:
        utils.calc_psms(df1, df1_f2)

        df_proteins, norm = utils.get_proteins_dataframe(df1_f2, decoy_prefix=args['prefix'],
            decoy_infix=args['infix'], all_decoys_2=decoy2, path_to_fasta=path_to_fasta, pif_threshold=args['pif_threshold'])
        tagnames = utils.get_tag_names(df1_f2.columns)

        logger.debug('Normalizing tag intensities by: %s', norm)
        logger.debug('Channels: %s', tagnames)

        nonnormalized = ['raw_' + n for n in tagnames]
        df1[nonnormalized] = df1[tagnames]
        df1[tagnames] /= norm
        df1_f2[tagnames] /= norm
        df1_peptides = df1.sort_values(key, ascending=True).drop_duplicates(['peptide'])
        df1_peptides_f = utils.filter_custom(df1_peptides[~df1_peptides['decoy1']], fdr=outfdr,
            key=key, is_decoy='decoy2', reverse=False, remove_decoy=False, ratio=pep_ratio, formula=1,
            correction=correction, loglabel='peptides')

        prot_ratio = 0.5
        df_proteins = df_proteins[~df_proteins['decoy1']]
        df_proteins_f = utils.filter_custom(df_proteins, fdr=outfdr, key='score', is_decoy='decoy2',
            reverse=False, remove_decoy=True, ratio=prot_ratio, formula=1, correction=correction, loglabel='proteins')
        utils.add_protein_groups(df_proteins_f, args['ms1'])
        df_protein_groups = df_proteins_f[df_proteins_f['groupleader']]

        logger.info('Final results at %s%% FDR level:', args['fdr'])
        logger.info('Identified PSMs: %s', df1_f2[~df1_f2['decoy2']].shape[0])
        logger.info('Identified peptides: %s', df1_peptides_f[~df1_peptides_f['decoy2']].shape[0])
        logger.info('Identified proteins: %s', df_proteins_f.shape[0])
        logger.info('Identified protein groups: %s', df_protein_groups.shape[0])
        logger.info('Processing finished.')

        return df1_peptides, df1_peptides_f, df_proteins, df_proteins_f, df_protein_groups
    else:
        logger.warning('PSMs cannot be filtered at %s%% FDR. Please increase allowed FDR.', args['fdr'])
        return (None,) * 5


def write_tables(outfolder, outbasename, df1, df1_f2, df1_peptides_f, df_proteins_f, df_protein_groups):
    df1.to_csv(utils.filename(outfolder, outbasename, 'psm_full'),
        sep='\t', index=False, columns=utils.get_columns_to_output(df1.columns, 'psm_full'))
    df1_f2[~df1_f2['decoy2']].to_csv(utils.filename(outfolder, outbasename, 'psm'),
        sep='\t', index=False, columns=utils.get_columns_to_output(df1_f2.columns, 'psm'))
    df1_peptides_f[~df1_peptides_f['decoy2']].to_csv(utils.filename(outfolder, outbasename, 'peptide'),
        sep='\t', index=False, columns=utils.get_columns_to_output(df1_peptides_f.columns, 'peptide'))

    df_proteins_f.to_csv(utils.filename(outfolder, outbasename, 'protein'), sep='\t', index=False,
            columns=utils.get_columns_to_output(df_proteins_f.columns, 'protein'))
    df_protein_groups.to_csv(utils.filename(outfolder, outbasename, 'protein_group'), sep='\t', index=False,
            columns=utils.get_columns_to_output(df_protein_groups.columns, 'protein'))
    logger.info('Output tables saved.')


def process_file(args, decoy2=None):
    """Run Scavager for a single file (`args['file']` should be a single file name).

    Parameters
    ----------

    args : dict
        A dictionary of parameters as produced from argparse in :py:func:`search.run`.
    decoy2 : set, optional
        A set of proteins labeled as "decoy2"

    Returns
    -------
    out : int
        Exit code. 0 for success, 1 for empty result, negative for errors.
    """
    fname = args['file']
    outfolder = utils.get_output_folder(args['output'], fname)
    outbasename = utils.get_output_basename(fname, args['name_suffix'])
    outfdr = args['fdr'] / 100
    correction = False if args['no_correction'] else (args['force_correction'] or None)
    logger.debug('Operating with correction = %s', correction)
    decoy_prefix = args['prefix']
    decoy_infix = args['infix']
    sf = args['separate_figures']
    logger.info('Loading file %s...', os.path.basename(fname))
    if args['enzyme']:
        cleavage_rule = utils.convert_tandem_cleave_rule_to_regexp(args['enzyme'])
    else:
        cleavage_rule = None

    try:
        df1, all_decoys_2 = utils.prepare_dataframe(fname, decoy_prefix=decoy_prefix,
            decoy_infix=decoy_infix, cleavage_rule=cleavage_rule, fdr=outfdr, decoy2set=decoy2)
    except utils.NoDecoyError:
        logger.error('No decoys were found. Please check decoy_prefix/infix parameter or your search output.')
        return -12
    except utils.WrongInputError:
        logger.error('Unsupported input file format. Use .pep.xml or .mzid files.')
        return -2
    except utils.EmptyFileError:
        logger.error('Input file %s is empty.', fname)
        return 1

    try:
        allowed_peptides, group_prefix, group_infix, group_regex = utils.variant_peptides(
            args['allowed_peptides'], args['group_prefix'], args['group_infix'], args['group_regex'])
    except ValueError as e:
        logger.error(e.args[0])
        return -3

    try:
        df1, df1_f2 = filter_dataframe(df1, outfdr, correction,
            allowed_peptides, group_prefix, group_infix, group_regex, decoy_prefix, decoy_infix)
    except CatBoostError as e:
        logger.error('There was an error in Catboost: %s', e.args)
        return -11

    if df1_f2 is None:
        return 1
    df1_peptides, df1_peptides_f, df_proteins, df_proteins_f, df_protein_groups = build_output_tables(
        df1, df1_f2, all_decoys_2, args)
    if df1_peptides is None:
        return 1

    write_tables(outfolder, outbasename, df1, df1_f2, df1_peptides_f, df_proteins_f, df_protein_groups)
    if args['create_pepxml']:
        pepxmltk.easy_write_pepxml([args['file']], utils.filename(outfolder, outbasename, 'pepxml'),
            set(df1_f2.loc[~df1_f2['decoy2'], 'spectrum']))

    if df1_f2[~df1_f2['decoy2']].shape[0] >= 3:
        plot_outfigures(df1, df1_f2[~df1_f2['decoy2']], df1_peptides, df1_peptides_f[~df1_peptides_f['decoy2']],
            outfolder, outbasename, df_proteins=df_proteins,
            df_proteins_f=df_proteins_f[~df_proteins_f['decoy2']], separate_figures=sf)
    return 0
